/*

*/

import tree.*;
import tokenizer.*;
import treelanguage.*;
import language2.*;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.nio.charset.Charset;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.util.List;

/**
   This program implements an "interpreter" for Language_2.
   <p>
   This program provides both a REPL (Read Eval Print Loop)
   for the language and a "script file" interpreter.
*/
public class Language_2
{
   public static void main(String[] args) throws IOException
   {
      // IMPORTANT: Set this to 0 or 1 depending on whether you need
      // to see all of the interpreter's debugging information.
      EvaluateAST.DEBUG = 1;

      if (args.length > 1)
      {
         System.out.println("Usage: Language_2 <script>");
      }
      else if (args.length == 1)
      {
         runFile(args[0]);
      }
      else
      {
         runPrompt();
      }
   }


   private static void runFile(final String path) throws IOException
   {
      final byte[] bytes = Files.readAllBytes(Paths.get(path));

      final String source = new String(bytes, Charset.defaultCharset());

      // Tokenize, parse, and evaluate the source file.
      try
      {
         final Tokenizer tokens = new Tokenizer(source);

         // Print the token list.
         System.out.println(tokens);

         final Tree<Token> ast = Parse2AST.parse(tokens);

         // Print the AST as an S-expression.
         System.out.println( EvaluateAST.toString(ast) + "\n" );

         // Pretty-print the abstract syntax tree.
         System.out.println( PrettyPrintAST.prettyPrint( ast ) + "\n" );

         // Print the infix version of the expression.
         System.out.println( AST2infix.ast2infix( ast ) + "\n" );

         // Create dot and png files from the AST.
         final String baseName = path;
         Tree2dot.tree2dot(ast, baseName);
         Tree2png.tree2png(baseName);

         // Evaluate the AST.
         final Value value = EvaluateAST.eval( ast );

         // Print the result.
         if (EvaluateAST.DEBUG > 0)
         {
            System.out.println("result = " + value);
         }
         else
         {
            System.out.println( "result = " + value.toSimpleString() );
         }
      }
      catch (TokenizeError e)
      {
         System.out.println(e);
      }
      catch (ParseError e)
      {
         System.out.println(e);
      }
      catch (EvalError e)
      {
         System.out.println(e);
      }
   }


   private static void runPrompt() throws IOException
   {
      final InputStreamReader input = new InputStreamReader(System.in);
      final BufferedReader reader = new BufferedReader(input);

      // Create an environment for the REPL's variable definitions.
      final Environment globalEnv = new Environment();

      System.out.println("Language 2.");

      // REPL (read, eval, print loop).
      for (int i = 1;  ; ++i)
      {
         System.out.print("> ");

         // Read an expression.
         final String source = reader.readLine();

         try
         {
            final Tokenizer tokens = new Tokenizer(source);
            if ( ! tokens.hasToken() )
            {
               continue; // skip over empty string
            }
            final Tree<Token> ast = Parse2AST.parse(tokens);

            if ( tokens.hasToken() )  // there shouldn't be any more tokens
               throw new ParseError("unexpected input: "
                                  + tokens.peek().lexeme
                                  + " at line " + tokens.peek().line
                                  + ", position " + tokens.peek().position
                                   );

            // Create dot and png files from the AST.
            final String baseName = String.format("repl.%02d", i);
            Tree2dot.tree2dot(ast, baseName);
            Tree2png.tree2png(baseName);

            // Evaluate the AST using this REPL's instance of the Environment.
            final Value value = EvaluateAST.evaluateProg( ast, globalEnv );

            // Print the result.
            if (EvaluateAST.DEBUG > 0)
            {
               // The "reverse prompt" idea comes from
               // the JavaScript repl in Chrome DevTools.
               System.out.println( "< " + value );
            }
            else
            {
               System.out.println( "< " + value.toSimpleString() );
            }
         }
         catch (TokenizeError e)
         {
            System.out.println(e);
         }
         catch (ParseError e)
         {
            System.out.println(e);
         }
         catch (EvalError e)
         {
            System.out.println(e);
         }
      }
   }
}
